from django.contrib.postgres.fields import JSONField
from django.utils.translation import gettext_lazy as _

from dateutil.relativedelta import relativedelta

from . import extract_relativedelta_args, forms


class RelativeDeltaStoreField(JSONField):
    description = _('A dateutil.relativedelta.relativedelta object')

    def formfield(self, **kwargs):
        return super().formfield(**{
            'form_class': forms.RelativeDeltaStoreField,
            **kwargs,
        })

    def from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return relativedelta(**value)

    def get_prep_value(self, value):
        if isinstance(value, relativedelta):
            value = extract_relativedelta_args(value)
        return super().get_prep_value(value)

    def validate(self, value, model_instance):
        if isinstance(value, relativedelta):
            value = extract_relativedelta_args(value)
        super().validate(value, model_instance)



from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from dateutil.relativedelta import relativedelta

from . import create_relativedelta_string, parse_relativedelta_string


class RelativeDeltaStoreField(forms.Field):
    def __init__(
            self, *, required=True, widget=None, label=None, initial=None,
            help_text=None, error_messages=None, show_hidden_initial=False,
            validators=(), localize=False, disabled=False, label_suffix=None,
            serializer=create_relativedelta_string,
            deserializer=parse_relativedelta_string
    ):
        self.serializer = serializer
        self.deserializer = deserializer

        super().__init__(
            required=required, widget=widget, label=label, initial=initial,
            help_text=help_text, error_messages=error_messages, show_hidden_initial=show_hidden_initial,
            validators=validators, localize=localize, disabled=disabled, label_suffix=label_suffix,
        )

    def prepare_value(self, value):
        if isinstance(value, (relativedelta, dict)):
            value = self.serializer(value)
        return value

    def to_python(self, value):
        if value in self.empty_values:
            return None
        else:
            return str(value).strip()

    def clean(self, value):
        value = super().clean(value)
        if value is None:
            return value
        relativedelta_kwargs, remainder = self.deserializer(value)
        if remainder:
            raise ValidationError(_('Invalid value at: %(remainder)s'), params={'remainder': remainder})
        return relativedelta(**relativedelta_kwargs)



import re
from typing import Optional, Tuple, Union

from dateutil.relativedelta import relativedelta


RELATIVE_UNITS = ('years', 'months', 'weeks', 'days', 'hours', 'minutes', 'seconds', 'microseconds')
ABSOLUTE_UNITS = ('year', 'month', 'day', 'hour', 'minute', 'second', 'microsecond')
SUPPORTED_UNITS = ABSOLUTE_UNITS + RELATIVE_UNITS

UNIT_PATTERN = re.compile(r'(?P<value>[+\-]?\d+)\s*(?P<unit>{units})\b'.format(units='|'.join(SUPPORTED_UNITS)))


def relativedelta_has_value_for(obj: relativedelta, attr: str) -> bool:
    value = getattr(obj, attr)
    if attr == 'weeks':  # calculated property
        return False
    if attr in RELATIVE_UNITS:
        return value != 0
    return value is not None


def extract_relativedelta_args(obj: relativedelta) -> dict:
    return dict((
        (unit, getattr(obj, unit)) for unit in SUPPORTED_UNITS if relativedelta_has_value_for(obj, unit)
    ))


# custom string (de-)serializer

def format_value(value, unit):
    format_str = '{value}{unit}' if unit in ABSOLUTE_UNITS else '{value:+d}{unit}'
    return format_str.format(value=value, unit=unit)


def create_relativedelta_string(value: Union[relativedelta, dict]) -> str:
    if isinstance(value, relativedelta):
        value = extract_relativedelta_args(value)
    sorted_units = sorted(value.keys(), key=lambda x: SUPPORTED_UNITS.index(x))
    return ' '.join([format_value(value=value[key], unit=key) for key in sorted_units])


def parse_relativedelta_string(value: str) -> Tuple[Optional[dict], Optional[str]]:
    remainder = value.strip()
    relativedelta_kwargs = dict()

    while True:
        match = UNIT_PATTERN.match(remainder)
        if not match:
            return None, remainder

        remainder = remainder[match.end('unit'):].strip()  # remove current match & strip whitespace
        relativedelta_kwargs[match.group('unit')] = int(match.group('value'))

        if not remainder:
            break

    return relativedelta_kwargs, remainder
